from posixpath import dirname
from textwrap import indent
from typing import *
from susc import File
from susc.things import *
from os import makedirs, path, write
from susc import log
from colorama import Fore
from nanoid import generate as nanoid

LICENSE = """ * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the “Software”), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""

def snake_to_pascal(name: str) -> str:
    words = name.split("_")
    return "".join(w[0].upper() + w[1:] for w in words)
def snake_to_camel(name: str) -> str:
    pascal = snake_to_pascal(name)
    return pascal[0].lower() + pascal[1:]

def regex_flags(pattern: re.Pattern):
    fstr = ""
    if pattern.flags & re.I: fstr += "i"
    if pattern.flags & re.M: fstr += "m"
    if pattern.flags & re.S: fstr += "s"
    return fstr
def type_validators(type_: SusType):
    vals = []
    for validator in type_.validators:
        val = validator.param + ": "
        restr = validator.restriction
        if isinstance(restr, re.Pattern): val += f"/{restr.pattern}/{regex_flags(restr)}"
        elif isinstance(restr, int): val += str(restr)
        elif isinstance(restr, str): val += restr
        elif isinstance(restr, range): val += f"[{restr[0]}, {restr[-1]}]"
        vals.append(val)
    if len(vals) == 0:
        return "{}"
    return "{ " + ', '.join(vals) + " }"
def type_to_amogus(type_: SusType, obj_types: Dict[str, str]) -> str:
    if type_.name == "List":
        elements = type_to_amogus(type_.args[0], obj_types)
        return f"new amogus.repr.List({elements}, {type_.args[1]}, {type_validators(type_)})"
    if type_.name == "Entity":
        return "new amogus.repr.Entity()"

    if type_.name in obj_types:
        # determine the type of the object from its name
        t_name = {
            "entities": "Entity",
            "bitfields": "EnumOrBf",
            "enums": "EnumOrBf",
            "confirmations": "Confirmation",
            "compounds": "Compound"
        }[obj_types[type_.name]]

        if t_name == "Entity":
            return "new amogus.repr.Entity()"

        if t_name == "Compound":
            return f"new amogus.repr.FieldArray({type_.name}Spec)"

        if t_name == "EnumOrBf":
            return f"new amogus.repr.EnumOrBf<{type_.name}>({type_.name}_SIZE)"
        
        return f"new amogus.repr.{t_name}({type_.name})"

    name = type_.name
    if name == "Int" and type_.args[0] > 4:
        name = "BigInteger"

    return f"new amogus.repr.{name}({', '.join([str(x) for x in type_.args] + [''])}{type_validators(type_)})"

def write_output(root_file: File, target_dir: str) -> None:
    proj_name = path.splitext(path.basename(root_file.path))[0]
    header = ("/* Generated by AMOGUS SUSC (https://github.com/amogus-api/susc)\n"
    f" * Project name: {proj_name}\n *\n"
    "" + LICENSE + "\n */\n\n")

    # pick random identifier
    proj_id = f"{proj_name}-{nanoid(size=10)}"

    # construct a name-to-type mapping
    objs = {}
    for thing in root_file.things:
        name = type(thing).__name__[3:].lower() + "s"
        if name == "entitys": # correct plural form
            name = "entities"
        objs[thing.name] = name

    with open(path.join(target_dir, "index.ts"), "w") as f:
        f.write(header)
        f.write("import * as amogus from \"amogus-driver\";\n\n")

        # write compounds
        compounds = [t for t in root_file.things if isinstance(t, SusCompound)]
        for thing in compounds:
            write_docstr(f, thing)
            f.write(f"export const {thing.name}Spec = {'{'}\n")
            write_field_array(f, thing.fields, objs, 1)
            f.write("};\n\n")

        # write enums
        enums_and_bfs = [t for t in root_file.things if isinstance(t, (SusEnum, SusBitfield))]
        for thing in enums_and_bfs:
            write_docstr(f, thing)
            f.write(f"export enum {thing.name} {'{'}\n")
            for member in thing.members:
                write_docstr(f, member, 1)
                prefix = "" if isinstance(thing, SusEnum) else "1 << "
                f.write(f"\t{member.name} = {prefix}{member.value},\n")
            f.write("}\n")
            f.write(f"export const {thing.name}_SIZE = {thing.size};\n\n\n")

        # write confirmations
        confirmations = [t for t in root_file.things if isinstance(t, SusConfirmation)]
        for conf in confirmations:
            # write spec
            name = snake_to_pascal(conf.name)
            f.write(f"const {name}Spec = {'{'}\n")
            f.write("\trequest: {\n")
            write_field_array(f, conf.req_parameters, objs)
            f.write("\t},\n")
            f.write("\tresponse: {\n")
            write_field_array(f, conf.resp_parameters, objs)
            f.write("\t}\n")
            f.write("};\n")
            # write class
            write_docstr(f, conf)
            f.write(f"export class {name} extends amogus.Confirmation<typeof {name}Spec> {'{'}\n")
            f.write("\tconstructor() {\n")
            f.write(f"\t\tsuper({name}Spec, {conf.value});\n")
            f.write("\t}\n")
            f.write("}\n")

        # write global methods
        methods = [t for t in root_file.things if isinstance(t, SusMethod)]
        for method in methods:
            # write spec
            name = snake_to_pascal(method.name)
            f.write(f"const {name}Spec = {'{'}\n")
            f.write(f"\tname: \"{method.name}\" as const,\n")
            if method.rate_limit:
                f.write(f"\trateLimit: [{method.rate_limit[0]}, {method.rate_limit[1]}] as const,\n")
            f.write("\tparams: {\n")
            write_field_array(f, method.parameters, objs)
            f.write("\t},\n")
            f.write("\treturns: {\n")
            write_field_array(f, method.returns, objs)
            f.write("\t},\n")
            conf_names = ", ".join(f"new {snake_to_pascal(conf)}()" for conf in method.confirmations)
            f.write(f"\tconfirmations: [{conf_names}]\n")
            f.write("};\n")
            # write class
            write_docstr(f, method)
            f.write(f"export class {name} extends amogus.Method<typeof {name}Spec> {'{'}\n")
            f.write("\tconstructor() {\n")
            f.write(f"\t\tsuper({name}Spec, {method.value}, undefined);\n")
            f.write("\t}\n")
            f.write("}\n")
            # write function
            write_docstr(f, method)
            f.write(f"async function {snake_to_camel(method.name)}(\n")
            f.write("\tthis: any | amogus.BoundSession,\n")
            f.write(f"\tparams: amogus.repr.FieldValue<typeof {name}Spec[\"params\"]>,\n")
            f.write(f"\tconfirm?: amogus.ConfCallback<{name}>,\n")
            f.write("\tsession?: amogus.Session<amogus.SpecSpace>\n")
            f.write(f"): Promise<amogus.repr.FieldValue<typeof {name}Spec[\"returns\"]>> {'{'}\n")
            f.write(f"\tconst method = new {name}();\n")
            f.write(f"\tmethod.params = params;\n")
            f.write(f"\treturn await (session ?? this.$session).invokeMethod(method, confirm);\n")
            f.write("}\n\n\n")

        # write entities
        entities = [t for t in root_file.things if isinstance(t, SusEntity)]
        for entity in entities:
            # write method specs and classes
            for method in entity.methods:
                name = f"{entity.name}_{snake_to_pascal(method.name)}"
                f.write(f"const {name}Spec = {'{'}\n")
                f.write(f"\tname: \"{entity.name}.{method.name}\" as const,\n")
                if method.rate_limit:
                    f.write(f"\trateLimit: [{method.rate_limit[0]}, {method.rate_limit[1]}] as const,\n")
                f.write("\tparams: {\n")
                write_field_array(f, method.parameters, objs)
                f.write("\t},\n")
                f.write("\treturns: {\n")
                write_field_array(f, method.returns, objs)
                f.write("\t},\n")
                conf_names = ", ".join(f"new {snake_to_pascal(conf)}()" for conf in method.confirmations)
                f.write(f"\tconfirmations: [{conf_names}]\n")
                f.write("};\n")
                write_docstr(f, method)
                f.write(f"export class {name} extends amogus.Method<typeof {name}Spec> {'{'}\n")
                f.write("\tconstructor() {\n")
                f.write(f"\t\tsuper({name}Spec, {method.value + (128 if method.static else 0)}, {entity.value});\n")
                f.write("\t}\n")
                f.write("}\n")

            # write spec
            name = entity.name
            f.write(f"const {name}Spec = {'{'}\n")
            f.write("\tfields: {\n")
            write_field_array(f, entity.fields, objs)
            f.write("\t},\n")
            f.write("\tmethods: {\n")
            for method in entity.methods:
                val = method.value + (128 if method.static else 0)
                f.write(f"\t\t{val}: new {entity.name}_{snake_to_pascal(method.name)}(),\n")
            f.write("\t}\n")
            f.write("};\n")
            # write class
            write_docstr(f, entity)
            f.write(f"export class {name} extends amogus.Entity<typeof {name}Spec> {'{'}\n")
            f.write("\tprotected static readonly session?: amogus.Session<amogus.SpecSpace>;\n")
            f.write("\treadonly dynSession?: amogus.Session<amogus.SpecSpace>;\n\n")
            f.write(f"\tconstructor(value?: amogus.repr.FieldValue<typeof {name}Spec[\"fields\"]>) {'{'}\n")
            f.write(f"\t\tsuper({name}Spec, {entity.value}, value);\n")
            f.write("\t}\n")

            # write fields
            f.write("\n")
            for field in entity.fields:
                f.write(f"\tget {field.name}() {'{'} return this.value?.{field.name}; {'}'}\n")

            # write method functions
            for method in entity.methods:
                name = f"{entity.name}_{snake_to_pascal(method.name)}"
                f.write("\n")
                write_docstr(f, method, 1)
                static = "static " if method.static else ""
                protected = "protected " if method.name == "get" else ""
                f.write(f"\t{protected}{static}async {snake_to_camel(method.name)}(\n")
                f.write(f"\t\tparams: amogus.repr.FieldValue<typeof {name}Spec[\"params\"]>,\n")
                f.write(f"\t\tconfirm?: amogus.ConfCallback<{name}>,\n")
                f.write("\t\tsession?: amogus.Session<amogus.SpecSpace>\n")
                f.write(f"\t): Promise<amogus.repr.FieldValue<typeof {name}Spec[\"returns\"]>> {'{'}\n")
                f.write(f"\t\tconst method = new {name}();\n")
                f.write(f"\t\tmethod.params = params;\n")
                if method.static:
                    f.write("\t\treturn await (session ?? this.session)!.invokeMethod(method, confirm);\n")
                else:
                    f.write("\t\tif(!this.value) throw new Error(\"Entity must have a value\");\n")
                    f.write("\t\tmethod.entityId = this.value.id;\n")
                    f.write("\t\treturn await (session ?? this.dynSession)!.invokeMethod(method, confirm);\n")
                f.write("\t}\n")

            # write $get()
            f.write("\n\tstatic async $get(id: bigint) {\n")
            f.write(f"\t\treturn (await this.get({'{'} id {'}'})).entity as amogus.ValuedEntity<{entity.name}>;\n")
            f.write("\t}\n")

            f.write("}\n\n\n")

        # write spec space
        f.write("\nexport function $specSpace(session: amogus.Session<amogus.SpecSpace>) {\n")
        f.write("\treturn {\n")
        f.write("\t\tspecVersion: \"2\" as const,\n")
        f.write(f"\t\tproject: \"{proj_id}\" as const,\n")
        f.write("\t\tglobalMethods: {\n")
        for method in methods:
            f.write(f"\t\t\t{method.value}: new {snake_to_pascal(method.name)}(),\n")
        f.write("\t\t},\n")
        f.write("\t\tentities: {\n")
        for entity in entities:
            f.write(f"\t\t\t{entity.value}: new class extends {entity.name} {'{'}\n")
            f.write("\t\t\t\treadonly dynSession = session;\n")
            f.write("\t\t\t\tprotected static readonly session = session;\n")
            f.write("\t\t\t} (),\n")
        f.write("\t\t},\n")
        f.write("\t\tconfirmations: {\n")
        for confirmation in confirmations:
            f.write(f"\t\t\t{confirmation.value}: new {confirmation.name}(),\n")
        f.write("\t\t},\n")
        f.write("\t};\n")
        f.write("}\n\n\n")

        # write $bind()
        f.write("\nexport function $bind(session: amogus.Session<amogus.SpecSpace>) {\n")
        f.write(f"\tif(session.specSpace.project !== \"{proj_id}\")\n")
        f.write("\t\tthrow new Error(\"failed to $bind: project identifier does not match\")\n\n")
        f.write("\treturn {\n")
        f.write("\t\t$session: session,\n")
        f.write("\t\t$close: async () => await session.stop(),\n")
        f.write("\t\t/*** METHODS ***/\n\n")
        for method in methods:
            write_docstr(f, method, 2)
            f.write(f"\t\t{snake_to_camel(method.name)},\n")
        f.write("\n\t\t/*** ENTITIES ***/\n\n")
        for entity in entities:
            write_docstr(f, entity, 2)
            f.write(f"\t\t{entity.name}: session.specSpace.entities[{entity.value}].constructor as typeof {entity.name},\n")
        f.write("\n\t\t/*** ENUMS AND BITFIELDS ***/\n\n")
        for thing in enums_and_bfs:
            write_docstr(f, thing, 2)
            f.write(f"\t\t{thing.name},\n")
        f.write("\t};\n")
        f.write("}\n")


def write_field_array(f, fields, objs, indent=2):
    indent = "\t" * indent

    f.write(f"{indent}required: {'{'}\n")
    for field in [f for f in fields if f.optional is None]:
        write_docstr(f, field, len(indent) + 1)
        f.write(f"{indent}\t{field.name}: {type_to_amogus(field.type_, objs)},\n")
    f.write(f"{indent}{'}'},\n")

    f.write(f"{indent}optional: {'{'}\n")
    for field in [f for f in fields if f.optional is not None]:
        write_docstr(f, field, len(indent) + 1)
        type_ = type_to_amogus(field.type_, objs)
        repr_class = type_[:type_.find("(")]
        f.write(f"{indent}\t{field.name}: [{field.optional}, {type_}] as const,\n")
    f.write(f"{indent}{'}'}\n")


def write_docstr(f, thing: SusThing, indent=0):
    indent = "\t" * indent
    if not thing.docstring:
        return

    if thing.docstring.count("\n") == 0:
        f.write(f"{indent}// {thing.docstring}\n")
        return

    f.write(f"{indent}/*\n")
    for line in thing.docstring.split("\n"):
        f.write(f"{indent} * {line}\n")
    f.write(f"{indent} */\n")


import argparse
from os import path
from colorama import Fore
from time import time

from . import File
from . import exceptions
from . import log
from . import lang_server

def highlight(file):
    for line in file.readlines():
        print(log.highlight_syntax(line), end='')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("source", help="file(s) to compile", type=argparse.FileType(mode="r", encoding="utf8"), nargs="*")
    parser.add_argument("-o", "--output", help="override output dir")
    parser.add_argument("-l", "--lang", help="override `set output` directive")
    parser.add_argument("-v", "--verbose", help="verbose logging", action="store_true")
    parser.add_argument("-p", "--highlight", help="print contents of a SUS file with highlighting", action="store_true")
    parser.add_argument("-e", "--single-line-errors", help="output errors in a parsable format", action="store_true")
    parser.add_argument("-s", "--language-server", help="run as a language server", action="store_true")
    parser.add_argument("-i", "--ls-stdio", help="run LS in stdio mode", action="store_true")
    args = parser.parse_args()

    exceptions.SINGLE_LINE_ERRORS = args.single_line_errors
    log.VERBOSE = args.verbose
    if log.VERBOSE:
        log.verbose("Verbose mode enabled")

    if args.language_server:
        lang_server.start(args.ls_stdio)
        return
    if args.ls_stdio:
        log.error("--ls-stdio can only be used with --language-server")
        return

    if args.highlight:
        for file in args.source:
            log.info(file.name)
            highlight(file)
        return

    if len(args.source) > 1 and args.output is not None:
        log.warn(f"More than one project specified. {Fore.GREEN}'-o {args.output}'{Fore.WHITE} will be ignored")
    
    if len(args.source) == 0:
        log.error("No source files specified")
        return

    successful = 0
    global_start = time()
    for i, source in enumerate(args.source):
        proj_start = time()
        if len(args.source) > 1:
            log.info(f"Compiling project {Fore.GREEN}'{source.name}'{Fore.WHITE} ({Fore.GREEN}{i + 1}/{len(args.source)}{Fore.WHITE})")

        sus_file = File()
        sus_file.load_from_file(source)

        # parse file and print diagnostics
        _, diagnostics = sus_file.parse()
        has_error = False
        for diag in diagnostics:
            exceptions.SourceError(diag).print()
            print()
            if diag.level == exceptions.DiagLevel.ERROR:
                has_error = True
        if has_error:
            continue
        
        langs = args.lang or sus_file.settings.get("output", None)
        if not langs:
            log.error(f"{Fore.RED}No output languages specified. Use the 'set output <language list>' directive in the root file or pass '-l <language list>' to the compiler")
            continue

        langs = langs.split()
        for lang in langs:
            output = args.output
            if len(args.source) > 1 or output is None:
                output = path.join(path.dirname(source.name), path.splitext(path.basename(source.name))[0] + "_output")
                
            try:
                sus_file.write_output(lang, path.join(output, lang))
            except exceptions.OutputError as ex:
                log.error(str(ex))
                continue

        proj_end = time()
        successful += 1
        took = int((proj_end - proj_start) * 1000)
        log.done(f"Compiled project into {len(langs)} language{'s' if len(langs) > 1 else ''} in {took}ms")

    global_end = time()
    took = int((global_end - global_start) * 1000)
    if len(args.source) > 1:
        log.done(f"Compiled {Fore.GREEN}{successful}/{len(args.source)}{Fore.WHITE} projects successfully in {took}ms")

if __name__ == "__main__":
    main()



from glob import iglob
from os import path
from pygls.server import LanguageServer
from pygls.lsp.methods import COMPLETION, TEXT_DOCUMENT_DID_CHANGE, HOVER
from pygls.lsp.types import (DidChangeTextDocumentParams, Diagnostic, Range,
                             Position, DiagnosticSeverity, CompletionOptions, CompletionParams,
                             CompletionList, CompletionItem, CompletionItemKind, HoverParams,
                             Hover, MarkedString)

from .things import SusBitfield, SusCompound, SusEntity, SusEnum, SusThing, SusField, SusType, SusValidator, SusMethod
from . import log
from . import File, KNOWN_SETTINGS

server = LanguageServer()
files: dict[str, File] = {}

@server.feature(TEXT_DOCUMENT_DID_CHANGE)
def did_change(ls: LanguageServer, params: DidChangeTextDocumentParams):
    path = params.text_document.uri[len("file://"):]

    global files
    ls.show_message_log("Parsing project")
    source = ls.workspace.get_document(params.text_document.uri).source

    file = files.get(params.text_document.uri, File())
    files[params.text_document.uri] = file

    file.load_from_text(source)
    file.path = path
    _, diagnostics = file.parse()
    diag_list = []

    for diag in diagnostics:
        for location in diag.locations:
            # ignore locations that are not in the current file
            if location.file.path != path:
                continue

            diag_list.append(Diagnostic(
                range=Range(
                    start=Position(line=location.line - 1, character=location.col - 1),
                    end=Position(line=location.line - 1, character=location.col - 1 + location.dur)
                ),
                message=diag.message,
                source="susc",
                severity=DiagnosticSeverity(diag.level.value)
            ))
        
    log.verbose("Pushing diagnostics")
    ls.publish_diagnostics(params.text_document.uri, diag_list)

    ls.show_message_log("Parsing done")

@server.feature(COMPLETION, CompletionOptions(trigger_characters=[":", "(", "[", ","]))
def completions(params: CompletionParams):
    global files
    file = files[params.text_document.uri]

    # what should we be finding?
    expected, state = file.insight(params.position.line, params.position.character)
    log.verbose(f"Parser expected {expected}")
    log.verbose(f"Parser state {state}")

    if "TYPE_IDENTIFIER" in expected:
        finding = "types"
    elif "VALIDATOR_IDENTIFIER" in expected:
        finding = "validators"
    elif "PATH" in expected:
        finding = "paths"
    elif "PARAMETER" in expected:
        finding = "parameters"
    else:
        return None

    # find precisely that
    log.verbose(f"Finding {finding}")
    items = []

    if finding == "types":
        for thing in file.things:
            if isinstance(thing, (SusEntity, SusCompound)):
                items.append(CompletionItem(label=thing.name, kind=CompletionItemKind.Class))
        items += [
            CompletionItem(label=n, kind=CompletionItemKind.TypeParameter)
            for n in ("Str", "Int", "List", "Bool")
        ]

    if finding == "validators":
        items += [
            CompletionItem(label=n, kind=CompletionItemKind.Property)
            for n in ["val", "len", "match"]
        ]

    if finding == "parameters":
        items += [
            CompletionItem(label=n, kind=CompletionItemKind.Property)
            for n in KNOWN_SETTINGS
        ]

    if finding == "paths":
        # find .sus files near this one
        for directory in file.search_paths():
            items += [
                CompletionItem(label=path.basename(n), kind=CompletionItemKind.File)
                for n in iglob(path.join(directory, "*.sus"))
            ]

    log.verbose("Sending completions")
    return CompletionList(
        is_incomplete=False,
        items=items
    )

def display_t_arg(a):
    if isinstance(a, SusType):
        return display_type(a)
    return str(a)

def display_t_val(v: SusValidator):
    return f"{v.param}: {v.restriction}"

def display_type(t: SusType):
    return f"{t.name}" +\
        "(" + ','.join(display_t_arg(a) for a in t.args) + ")" +\
        "[" + ','.join(display_t_val(v) for v in t.validators) + "]"

def display_field(field: SusField):
    opt = f"opt({field.optional}) " if field.optional != None else ""
    return f"{field.name}: {opt}{display_type(field.type_)}"

def display_method(method: SusMethod):
    kw = "staticmethod" if method.static else "method"
    return f"\t{kw} {method.name}({method.value})" + "{\n" +\
        "".join(f"\t\t{display_field(f)};\n" for f in method.parameters) +\
        "\t\treturns {\n" +\
        "".join(f"\t\t\t{display_field(f)};\n" for f in method.returns) +\
        "\t\t}\n" +\
        "\t}\n"

def display_thing(thing: SusThing):
    if isinstance(thing, (SusBitfield, SusEnum)):
        kw = "enum" if isinstance(thing, SusEnum) else "bitfield"
        return f"{kw} {thing.name} " + "{\n" +\
            "".join(f"\t{m.name}({f.value}),\n" for m in thing.members) +\
            "}"

    elif isinstance(thing, SusEntity):
        return f"entity {thing.name} " + "{\n" +\
            "".join(f"\t{display_field(f)};\n" for f in thing.fields) +\
            "".join("\n" + display_method(m) for m in thing.methods) +\
            "}"

    elif isinstance(thing, SusCompound):
        return f"compound {thing.name} " + "{\n" +\
            "".join(f"\t{display_field(f)};\n" for f in thing.fields) +\
            "}"

@server.feature(HOVER)
def hover(params: HoverParams):
    global files
    file = files[params.text_document.uri]

    # list of things that can be used as types
    things: list[SusThing] = []
    for thing in file.things:
        if isinstance(thing, (SusEntity, SusCompound, SusEnum, SusBitfield)):
            things.append(thing)

    line = file.source.split("\n")[params.position.line]
    # find first non-alpha char to the left
    start = params.position.character
    while start > 0 and line[start - 1].isalpha():
        start -= 1
    # find last alpha char to the right
    end = start
    while end < len(line) and line[end].isalpha():
        end += 1
    # find token that is being hovered over
    token = line[start:end]

    # find the thing that is being hovered over
    for thing in things:
        if thing.location.file != file:
            continue
        if isinstance(thing, (SusEntity, SusEnum, SusBitfield, SusCompound)):
            if thing.name == token:
                contents = [MarkedString(
                    language="sus",
                    value=display_thing(thing)
                )]
                if thing.docstring:
                    contents.append(MarkedString(
                        language="markdown",
                        value=thing.docstring
                    ))
                return Hover(contents=contents)
        
    if token in ("Str", "Int", "List", "Bool"):
        return Hover(contents=[MarkedString(
            language="sus",
            value = f"(built-in) {token}"
        )])

def start(io: bool):
    if io:
        log.ALL_STDERR = True
        server.start_io()
    else:
        log.done("Starting on localhost:9090")
        server.start_tcp("127.0.0.1", 9090)



from posixpath import dirname
from textwrap import indent
from typing import *
from susc import File
from susc.things import *
from os import makedirs, path
from shutil import copy, copytree, rmtree
from susc import log
from colorama import Fore
from markdown import markdown

LICENSE = """
\tCopyright © 2021 portasynthinca3

\tPermission is hereby granted, free of charge, to any person obtaining a copy of this software and
\tassociated documentation files (the “Software”), to deal in the Software without restriction,
\tincluding without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
\tand/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
\tsubject to the following conditions:

\tThe above copyright notice and this permission notice shall be included in all copies or substantial
\tportions of the Software.

\tTHE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
\tLIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
\tNO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
\tWHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
\tSOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE."""

def write_output(root_file: File, target_dir: str) -> None:
    proj_name = path.splitext(path.basename(root_file.path))[0]
    header = ("<!--\n\tGenerated by AMOGUS SUS (https://github.com/portasynthinca3/amogus)\n"
    f"\tProject name: {proj_name}\n\n"
    "" + LICENSE + "\n-->\n\n")

    # copy static presets
    copytree(path.join(dirname(__file__), "static"), path.join(target_dir, "static"))

    # copy topbar logo
    try:
        if "html_topbar_logo" in root_file.settings:
            topbar_logo = root_file.settings["html_topbar_logo"]
        else:
            log.info(f"Use {Fore.GREEN}'set html_topbar_logo <path>'{Fore.WHITE} to override the default topbar logo")
            topbar_logo = path.join(dirname(__file__), "topbar_default.png")
        copy(topbar_logo, path.join(target_dir, "topbar.png"))
    except:
        log.warn("Failed to copy topbar logo")

    if "html_topbar_title" in root_file.settings:
        topbar_title = root_file.settings["html_topbar_title"]
    else:
        log.info(f"Use {Fore.GREEN}'set html_topbar_title <path>'{Fore.WHITE} to override the default title")
        topbar_title = proj_name

    things = root_file.things

    log.verbose("Writing index.html")
    with open(path.join(target_dir, "index.html"), "w") as f:
        f.write(header)
        f.write("<html>\n")

        # head
        f.write("\t<head>\n")
        f.write(f"\t\t<title>{topbar_title} - Index</title>\n")
        f.write("\t\t<link rel='stylesheet' href='static/style.css'>\n")
        f.write("\t</head>\n")
        f.write("\t<body>\n")

        # topbar
        f.write("\t\t<div id='topbar'>\n")
        f.write("\t\t\t<img src='topbar.png' width='48' height='48'>\n")
        f.write(f"\t\t\t<a href='index.html'>{topbar_title}</a>\n")
        f.write("\t\t</div>\n")

        # thing lists
        enums = sorted([t for t in things if isinstance(t, SusEnum)], key=lambda t: t.name)
        methods = sorted([t for t in things if isinstance(t, SusMethod)], key=lambda t: t.name)
        entities = sorted([t for t in things if isinstance(t, SusEntity)], key=lambda t: t.name)
        bitfields = sorted([t for t in things if isinstance(t, SusBitfield)], key=lambda t: t.name)
        confirmations = sorted([t for t in things if isinstance(t, SusConfirmation)], key=lambda t: t.name)
        write_index_list(f, "Enums", "enum", enums)
        write_index_list(f, "Bitfields", "bf", bitfields)
        write_index_list(f, "Entities", "entity", entities)
        write_index_list(f, "Global methods", "gm", methods)
        write_index_list(f, "Confirmations", "conf", confirmations)

        # closing
        f.write("\t</body>\n")
        f.write("</html>\n")

    # create a page for each thing
    for thing in things:
        title, prefix = {
            SusEnum: ("enum", "enum_"),
            SusBitfield: ("bitfield", "bf_"),
            SusEntity: ("entity", "entity_"),
            SusMethod: ("global method", "gm_"),
            SusConfirmation: ("confirmation", "conf_")
        }[type(thing)]

        file_name = prefix + thing.name.lower() + ".html"
        log.verbose(f"Writing {file_name}")
        with open(path.join(target_dir, file_name), "w") as f:
            f.write(header)
            f.write("<html>\n")

            # head
            f.write("\t<head>\n")
            f.write(f"\t\t<title>{topbar_title} - {thing.name} {title}</title>\n")
            f.write("\t\t<link rel='stylesheet' href='static/style.css'>\n")
            f.write("\t</head>\n")
            f.write("\t<body>\n")

            # topbar
            f.write("\t\t<div id='topbar'>\n")
            f.write("\t\t\t<img src='topbar.png' width='48' height='48'>\n")
            f.write(f"\t\t\t<a href='index.html'>{topbar_title} <span><code>{thing.name}</code> {title}</span></a>\n")
            f.write("\t\t</div>\n")

            write_thing(f, thing)

            # closing
            f.write("\t</body>\n")
            f.write("</html>\n")


def format_docstring(docstr, indentation=2):
    if docstr is None:
        return ""
    docstr = docstr.replace("\n", "\n\n")
    return indent(f"<div class='docstring'>\n{markdown(docstr)}\n</div>\n", "\t" * indentation)
def format_type(type_: SusType):
    args = type_.args[:]
    for i, arg in enumerate(args):
        if isinstance(arg, int):
            args[i] = str(arg)
        elif isinstance(arg, SusType):
            args[i] = format_type(arg)
    return type_.name + "(" + ", ".join(args) + ")"

def write_field(f, field: SusField, title: str, css_class: str):
    f.write("\t\t<div class='thing-member'>\n")
    f.write(f"\t\t\t<h3>{title}: <code class='{css_class}'>{field.name}</code></h3>\n")
    f.write(format_docstring(field.docstring, indentation=3))
    f.write(f"\t\t\t<div class='thing-param'>Type: <code>{format_type(field.type_)}</code></div>\n")
    if field.caching:
        f.write(f"\t\t\t<div class='thing-param'>Caching: {field.caching}</div>\n")
    if field.optional is not None:
        f.write(f"\t\t\t<div class='thing-param'>Optional ID: {field.optional}</div>\n")
    if field.default is not None:
        f.write(f"\t\t\t<div class='thing-param'>Default value: <code>{field.default}</code></div>\n")
    f.write("\t\t</div>\n")

def transform_cond_list(l):
    return ", ".join(f"<code class='field'>{m}</code>" for m in l)
def write_method(f, method: SusMethod, write_header=True):
    if write_header:
        f.write("\t\t<div class='thing-member'>\n")
    value_elm = "div" if write_header else "h3"
    if write_header:
        f.write(f"\t\t\t<h3>{'static ' if method.static else ''}method: <code class='method'>{method.name}</code></h3>\n")
    f.write(format_docstring(method.docstring, indentation=3))

    f.write(f"\t\t\t<{value_elm} class='thing-param'>Value: {method.value}</{value_elm}>\n")
    if method.errors:
        f.write(f"\t\t\t<{value_elm} class='thing-param'>Errors: {transform_cond_list(method.errors)}</{value_elm}>\n")
    if method.states:
        f.write(f"\t\t\t<{value_elm} class='thing-param'>States: {transform_cond_list(method.states)}</{value_elm}>\n")
    if method.confirmations:
        f.write(f"\t\t\t<{value_elm} class='thing-param'>Confirmations: {transform_cond_list(method.confirmations)}</{value_elm}>\n")

    for param in method.parameters:
        write_field(f, param, "parameter", "param")
    for param in method.returns:
        write_field(f, param, "return value", "ret-val")

    if write_header:
        f.write("\t\t</div>\n")

def write_thing(f, thing: SusThing):
    if isinstance(thing, (SusEnum, SusBitfield)):
        f.write(format_docstring(thing.docstring))
        f.write(f"\t\t<h3 class='thing-param'>Size: {thing.size} bytes ({thing.size * 8} bits)</h3>\n")

        for field in sorted(thing.members, key=lambda m: m.value):
            f.write("\t\t<div class='thing-member'>\n")
            f.write(f"\t\t\t<h3>member: <code class='field'>{field.name}</code></h3>\n")
            f.write(format_docstring(field.docstring, indentation=3))
            f.write(f"\t\t\t<div class='thing-param'>Value: {field.value}</div>\n")
            f.write("\t\t</div>\n")
    
    elif isinstance(thing, SusEntity):
        f.write(format_docstring(thing.docstring))
        f.write(f"\t\t<h3 class='thing-param'>Value: {thing.value}</h3>\n")
        for field in thing.fields:
            write_field(f, field, "field", "field")
        for method in thing.methods:
            write_method(f, method)
    
    elif isinstance(thing, SusConfirmation):
        f.write(format_docstring(thing.docstring))
        f.write(f"\t\t<h3 class='thing-param'>Value: {thing.value}</h3>\n")
        for param in thing.req_parameters:
            write_field(f, param, "request parameter", "param")
        for param in thing.resp_parameters:
            write_field(f, param, "response parameter", "ret-val")

    elif isinstance(thing, SusMethod):
        write_method(f, thing, False)

def correct_noun_form(amt, text):
    if amt == 1:
        return str(amt) + " " + text
    else:
        return str(amt) + " " + text + "s"
def noun_list(lst):
    return ", ".join([correct_noun_form(amt, noun) for amt, noun in lst if amt > 0])
def write_index_list(f, title, href_pref, elements):
    if len(elements) == 0:
        return
    f.write("\t\t<ul class='index-section'>\n")
    f.write(f"\t\t\t<h1>{title}</h1>\n")

    for elm in elements:
        info = ""
        if isinstance(elm, (SusEnum, SusBitfield)):
            info = noun_list([
                (len(elm.members), "member")
            ])
        if isinstance(elm, SusConfirmation):
            info = noun_list([
                (len(elm.req_parameters), "req. parameter"),
                (len(elm.resp_parameters), "resp. parameter")
            ])
        if isinstance(elm, SusMethod):
            info = noun_list([
                (len(elm.parameters), "parameter"),
                (len(elm.returns), "return value")
            ])
        if isinstance(elm, SusEntity):
            static = [m for m in elm.methods if m.static]
            normal = [m for m in elm.methods if not m.static]
            info = noun_list([
                (len(elm.fields), "field"),
                (len(static), "static method"),
                (len(normal), "dynamic method")
            ])
        if len(info) > 0:
            info = "(" + info + ")"

        f.write(f"\t\t\t<li><a href='{href_pref}_{elm.name.lower()}.html'>{elm.name}</a><span>{info}</span></li>\n")

    f.write("\t\t</ul>\n")


from io import TextIOWrapper
from shutil import rmtree
from colorama.ansi import Fore
import lark
from lark.exceptions import UnexpectedInput
from os import path, makedirs
from importlib import import_module
from math import ceil

from .things import *
from . import log
from . import linker
from .exceptions import DiagLevel, Diagnostic, OutputError, SearchError, SourceError

KNOWN_SETTINGS = ["output", "html_topbar_logo", "html_topbar_title"]

# read the description file
with open(path.join(path.dirname(__file__), "sus.lark")) as f:
    lark_parser = lark.Lark(f.read(), parser="lalr")

def token_to_str(token: str):
    return {
        "LPAR": "'('", "RPAR": "')'",
        "LBRACE": "'{'", "RBRACE": "'}'",
        "LSQB": "'['", "RSQB": "']'",
        "DOCSTRING": "'@>'",
        "COLON": "':'",
        "SEMICOLON": "';'",
        "COMMA": "','",
        "PLUS": "'+'",

        "TYPE_IDENTIFIER": "name",
        "ROOT_IDENTIFIER": "name",
        "METHOD_IDENTIFIER": "name",
        "FIELD_IDENTIFIER": "field",
        "VALIDATOR_IDENTIFIER": "validator",
        "NUMBER": "number",
        "SIGNED_NUMBER": "signed_number",
        "REGEX": "regex",
        "RANGE": "range",
        "PARAMETER": "parameter",
        "VALUE": "value",
    # if none matched, turn TOKEN into 'token'
    }.get(token, "'" + token.lower() + "'")

class File():
    def __init__(self, parent=None, root=None):
        self.parent = parent
        self.root = root or self
        self.settings = {}
        self.dependencies = []
        self.things = []
        self.diagnostics = []
        if self.parent is None:
            self.all_loaded = set()

    def load_from_text(self, source, path=None):
        self.source = source
        self.path = path.abspath(path) or "<from source>"

        if self.parent is None:
            self.all_loaded.add(self.path)

        log.verbose(f"Loaded from source: {self.path} {Fore.LIGHTBLACK_EX}{'(root)' if not self.parent else ''}", "load")

    def load_from_file(self, source: str|TextIOWrapper):
        # read the file
        if isinstance(source, str):
            self.path = source
            source = open(source, "r")
        else:
            self.path = path.abspath(source.name)
        self.source = source.read()
        source.close()

        if self.parent is None:
            self.all_loaded.add(self.path)

        log.verbose(f"Loaded file: {self.path} {Fore.LIGHTBLACK_EX}{'(root)' if not self.parent else ''}", "load")

    def search_paths(self):
        return [
            "",
            ".",
            path.dirname(self.path), # next to this file
            path.join(path.dirname(__file__), "stdlib") # in the standard library
        ]

    def resolve_source(self, p):
        targets = [path.join(d, p) for d in self.search_paths()]
        targets += [t + ".sus" for t in targets if not t.endswith(".sus")] # try with .sus extension
        targets = [path.abspath(t) for t in targets]
        targets = sorted(set(targets), key=lambda x: targets.index(x)) # remove duplicates keeping the order
        # try each one
        for target in targets:
            try:
                return open(target, "r")
            except FileNotFoundError: pass

        locations = '\n'.join(targets)
        raise SearchError(f"Couldn't find or open '{p}' in any of the following locations:\n{locations}")

    # provides insight into the parser state at that point
    def insight(self, line: int, col: int) -> tuple[set[str], list]:
        # convert line and column numbers to position within the string
        pos = 0
        for _ in range(line):
            pos = self.source.find("\n", pos) + 1
        pos += col

        # get the source up to that point
        source = self.source[:pos]

        # try parsing
        try:
            lark_parser.parse(source)
        except UnexpectedInput as e:
            return e.expected, e.state.value_stack

        return None, None # no expected tokens here

    def __parsing_error(self, e: lark.UnexpectedToken):
        log.verbose(f"Parsing error: {e}", "corrector")

        parser = e.interactive_parser
        tok: Token = e.token
        token = tok.value.split(' ')[0]
        expected = ', '.join(token_to_str(t) for t in e.expected)
        error_text = f"Expected{' one of:' if len(e.expected) > 1 else ''} {expected}"
        dur = len(token)

        if token == "":
            # empty token = EOF
            line = self.source.split('\n')[e.line - 1]
            location = Location(self, e.line, len(line) + 1, 0)
        else:
            location = Location(self, e.line, e.column, dur)

        diag = Diagnostic([location], DiagLevel.ERROR, error_text)
        self.diagnostics += [diag]

        # inform the user about our naming conventions :)
        # while trying to rename
        inter = e.expected.intersection({"TYPE_IDENTIFIER", "ROOT_IDENTIFIER"})
        if len(inter):
            diag.message = "This identifier should use PascalCase"
            tok.type = inter.pop()
            tok.value = tok.value[0].upper() + tok.value[1:]
            log.verbose(f"Renamed '{token}' to '{tok.value}'", "corrector")
            parser.feed_token(tok)
            return True

        inter = e.expected.intersection({"FIELD_IDENTIFIER", "METHOD_IDENTIFIER", "VALIDATOR_IDENTIFIER"})
        if len(inter):
            diag.message = "This identifier should use snake_case"
            tok.type = inter.pop()
            tok.value = tok.value.lower()
            parser.feed_token(tok)
            return True

        # fill in missing semicolons and things
        fill_in = {
            "SEMICOLON": ";", "COLON": ":",
            "RPAR": ")", "RBRACE": "}", "RSQB": "]",
            "COMMA": ",",
        }
        for k, v in fill_in.items():
            if k in e.expected:
                log.verbose(f"Inserted {k}", "corrector")
                parser.feed_token(Token(k, v))

                # insert an additional semicolon if the next token is not it
                if k in {"RPAR", "RSQB"} and tok.type != "SEMICOLON":
                    parser.feed_token(Token("SEMICOLON", ";"))
                    log.verbose(f"Inserted SEMICOLON", "corrector")

                if tok.type == "ROOT_IDENTIFIER":
                    tok.type = "TYPE_IDENTIFIER"
                log.verbose(f"Inserted original token", "corrector")
                print(parser.parser_state.value_stack, tok.type, tok.value)
                parser.feed_token(tok)

                return True

        # fill in missing numeric values
        structure = ([None, None] + parser.parser_state.value_stack)[-2]
        if e.expected == {"LPAR"} and structure and structure.type in {"ENTITY", "GLOBALMETHOD", "METHOD", "STATICMETHOD", "CONFIRMATION"}:
            diag.message = "Missing numeric value"
            log.verbose(f"Inserted '(0)'", "corrector")
            parser.feed_token(Token("LPAR", "("))
            parser.feed_token(Token("NUMBER", "0"))
            parser.feed_token(Token("RPAR", ")"))
            parser.feed_token(tok)
            return True

        return False

    def parse(self) -> Tuple[list[SusThing], list[Diagnostic]]:
        log.verbose(f"Parsing {Fore.WHITE}{self.path}", "parser")
        self.things = []
        self.dependencies = []
        self.diagnostics = []

        try:
            self.tree = lark_parser.parse(self.source, on_error=self.__parsing_error)
            log.verbose(f"AST constructed", "parser")
        except UnexpectedInput as e:
            log.verbose("No corrections or invalid correction", "corrector")
            # parsing can't continue any further, just return
            return [], self.diagnostics

        # deconstruct the syntax tree
        for thing in self.tree.children:
            if thing.data == "inclusion":
                name = thing.children[0]
                log.verbose(f"Encountered inclusion:{Fore.LIGHTBLACK_EX} path={Fore.WHITE}{name}", "parser")
                # find dependency
                try:
                    source = self.resolve_source(name.value)
                except SearchError as ex:
                    return [], [Diagnostic([Location(self, name.line, name.column, len(name))],
                        DiagLevel.ERROR, ex.msg)]

                # load it
                if source.name not in self.root.all_loaded:
                    dependency = File(self, self.root)
                    dependency.load_from_file(source)
                    self.dependencies.append(dependency)
                    self.root.all_loaded.add(source.name)
                else:
                    return [], [Diagnostic([Location(self, name.line, name.column, len(name))],
                        DiagLevel.ERROR, "This file has already been included directly or by a dependency within this project\n" +\
                        f"Note: inclusion resolved to '{source.name}'")]

            elif thing.data == "setting":
                name = thing.children[0]
                value = thing.children[1]
                log.verbose(f"Encountered setting:{Fore.LIGHTBLACK_EX} name={Fore.WHITE}{name}{Fore.LIGHTBLACK_EX} value={Fore.WHITE}{value}", "parser")
                if name.value not in KNOWN_SETTINGS:
                    self.diagnostics.append(Diagnostic([Location(self, name.line, name.column, len(name))], DiagLevel.WARN, "Unknown setting"))
                self.settings[name.value] = value.value

            else:
                if thing.data == "definition":
                    thing = thing.children[0]
                log.verbose(f"AST subtree: {log.highlight_ast(thing)}", "parser")
                thing = convert_ast(thing, self)
                log.verbose(f"Converted AST subtree: {Fore.WHITE}{log.highlight_thing(thing)}", "parser")
                self.things.append(thing)
                # generate standard methods for entities
                if isinstance(thing, SusEntity):
                    thing.methods.append(SusMethod(
                        thing.location,
                        f"Gets {thing.name} by ID",
                        True,
                        "get",
                        127,
                        [SusField(thing.location, "ID of the entity to get", "id", SusType(thing.location, None, "Int", [8], []), None)],
                        [SusField(thing.location, "Entity with that ID", "entity", SusType(thing.location, None, thing.name, [], []), None)],
                        ["invalid_id"],
                        [],
                        None
                    ))
                    thing.methods.append(SusMethod(
                        thing.location,
                        f"Updates {thing.name}",
                        False,
                        "update",
                        127,
                        [SusField(thing.location, "The values to update", "entity", SusType(thing.location, None, thing.name, [], []), None)],
                        [],
                        ["invalid_entity"],
                        [],
                        None
                    ))

        log.verbose(f"Parsing dependencies for {Fore.WHITE}{self.path}", "deps")
        # parse dependencies
        things = self.things
        for dep in self.dependencies:
            things += dep.parse()[0]
        log.verbose(f"Parsing dependencies done for {Fore.WHITE}{self.path}", "deps")

        # run linker
        if not self.parent:
            things, diag = linker.run(things)
            self.diagnostics += diag

        self.things = things
        return things, self.diagnostics

    def write_output(self, lang, target_dir):
        if not self.things:
            raise OutputError("No data to write. Call parse() first")

        try:
            module = import_module(".output." + lang, package=__package__)
            target_dir = path.abspath(target_dir)

            # clean output dir
            try: rmtree(target_dir)
            except FileNotFoundError: pass
            try: makedirs(target_dir)
            except FileExistsError: pass

            # call the module
            module.write_output(self, target_dir)
        except ImportError as ex:
            log.verbose(ex)
            raise OutputError(f"No output package for language '{lang}' or it is broken")



from abc import ABC
from dataclasses import dataclass
from os import times
from typing import *
from enum import Enum
from lark.lexer import Token
from lark.tree import Tree
import re
from .exceptions import SourceError, Location
from .log import verbose
from textwrap import dedent

@dataclass
class SusThing(ABC):
    location: Location
    docstring: str

@dataclass
class SusValidator(SusThing):
    param: str
    restriction: Any

class SusTypeBase(SusThing):
    @overload
    def find_errors(self, identifiers):
        raise NotImplemented()

@dataclass
class SusType(SusTypeBase):
    name: str
    args: List[Any]
    validators: List[SusValidator]
    def find_errors(self, identifiers):
        if self.name == "Int":
            if len(self.args) != 1:
                return "Int requires one argument"
            if not isinstance(self.args[0], int) or self.args[0] <= 0:
                return "Argument to Int should be a positive integer"
            for v in self.validators:
                if v.param != "val":
                    return "The following validators are valid for Int: 'val'"
                if not isinstance(v.restriction, range):
                    return "Int[val] must be a range"
        elif self.name == "Str":
            if len(self.args) != 0:
                return "Str takes no arguments"
            for v in self.validators:
                if v.param not in ["len", "match"]:
                    return "The following validators are valid for Str: 'len', 'match'"
                if v.param == "len" and not isinstance(v.restriction, range):
                    return "Str[len] must be a range"
                if v.param == "match" and not isinstance(v.restriction, re.Pattern):
                    return "Str[match] must be a regular expression"
        elif self.name == "List":
            if len(self.args) != 2:
                return "List takes two arguments"
            if not isinstance(self.args[0], SusTypeBase):
                return "First argument to List should be a type"
            err = self.args[0].find_errors(identifiers)
            if err is not None: return err
            if not isinstance(self.args[1], int) or self.args[1] <= 0:
                return "Second argument to List should be a positive integer"
            for v in self.validators:
                if v.param != "cnt":
                    return "The following validators are valid for List: 'cnt'"
                if not isinstance(v.restriction, range):
                    return "List[cnt] must be a range"
        elif self.name == "Bool":
            if len(self.args) != 0:
                return "Bool takes no arguments"
            if len(self.validators) != 0:
                return "Bool can't be validated"
        elif self.name == "Bin":
            if len(self.args) != 0:
                return "Bin takes no arguments"
            for v in self.validators:
                if v.param != "len":
                    return "The following validators are valid for Bin: 'len'"
                if not isinstance(v.restriction, range):
                    return "Bin[len] must be a range"
        elif self.name not in identifiers:
            return "Unknown type"

@dataclass
class SusCompoundMember(SusThing):
    name: str
    type_: SusTypeBase

@dataclass
class SusCompound(SusTypeBase):
    members: List[SusCompoundMember]
    def find_errors(self, identifiers):
        for m in self.members:
            err = m.type_.find_errors(identifiers)
            if err != None:
                return err

@dataclass
class SusEnumMember(SusThing):
    name: str
    value: int

@dataclass
class SusEnum(SusThing):
    name: str
    size: int
    members: List[SusEnumMember]

@dataclass
class SusBitfield(SusThing):
    name: str
    size: int
    members: List[SusEnumMember]

@dataclass
class SusField(SusThing):
    name: str
    type_: SusType
    optional: int

@dataclass
class SusMethod(SusThing):
    static: bool
    name: str
    value: int
    parameters: List[SusField]
    returns: List[SusField]
    errors: List[str]
    confirmations: List[str]
    rate_limit: Tuple[int, int]

@dataclass
class SusEntity(SusThing):
    name: str
    value: int
    fields: List[SusField]
    methods: List[SusMethod]

@dataclass
class SusConfirmation(SusThing):
    name: str
    value: int
    req_parameters: List[SusField]
    resp_parameters: List[SusField]

@dataclass
class SusCompound(SusThing):
    name: str
    fields: List[SusField]


def convert_docstring(doc: str) -> str:
    if doc is None:
        return None
    doc = doc.value
    doc = doc.strip(" \r\n")
    doc = doc[2:-2] # remove @> and <@
    doc = doc.strip("\r\n")
    return dedent(doc).strip()

def convert_ebf_members(ast, file): # Enum or Bitfield
    list = []
    for member in ast:
        if not member: continue
        doc = convert_docstring(member.children[0])
        name = member.children[1]
        value = int(member.children[2].value)
        list.append(SusEnumMember(Location(file, name.line, name.column, len(name.value)), doc, name.value, value))
    return list
    
def convert_range(ast, max_val):
    if len(ast.children) == 1:
        return range(int(ast.children[0].value), max_val)
    elif len(ast.children) == 2:
        return range(int(ast.children[0].value), int(ast.children[1].value) + 1) # python ranges exclude the right end

def convert_type(ast, file):
    name = ast.children[0]
    args = []
    validators = []
    for directive in ast.children[1:]:
        if directive is None:
            continue
        elif directive.data == "type_argument":
            arg_val = directive.children[0]
            if isinstance(arg_val, Token) and arg_val.type == "NUMBER":
                args.append(int(arg_val.value))
            elif isinstance(arg_val, Tree) and arg_val.data in ["type", "compound"]:
                args.append(convert_type(arg_val, file))
        elif directive.data == "type_validator":
            val_name = directive.children[0] # validator name
            val_val = directive.children[1].children[0] # validator value
            if isinstance(val_val, Token) and val_val.type == "SIGNED_NUMBER":
                val_val = int(val_val)
            elif isinstance(val_val, Token) and val_val.type == "REGEX":
                regex_tokens = val_val.split('/')
                regex = '/'.join(regex_tokens[1:-1])
                flags = regex_tokens[-1]
                flags_num = 0
                for f in flags:
                    if f == "i": flags_num |= re.I
                    if f == "m": flags_num |= re.M
                    if f == "s": flags_num |= re.S
                try:
                    val_val = re.compile(regex, flags_num)
                except re.error as exc:
                    raise SourceError([Location(file, val_val.line, val_val.column + exc.pos + 1, 0)],
                        f"Invalid regular expression: {exc.msg}")
            elif isinstance(val_val, Tree) and val_val.data == "range":
                max_val = 2 ** 64
                # we can infer the maximum value from the int length
                if name == "Int" and len(args) == 1 and isinstance(args[0], int):
                    max_val = 2 ** (args[0] * 8)
                val_val = convert_range(val_val, max_val)
            validators.append(SusValidator(Location(file, val_name.line, val_name.column, len(val_name.value)), None,
                val_name.value, val_val))

    location = Location(file, name.line, name.column, len(name.value))
    return SusType(location, None, name.value, args, validators)

def convert_opt(ast):
    if ast is None:
        return ast
    return int(ast.value)

def convert_value(ast):
    if isinstance(ast, Tree) and ast.data == "value":
        ast = ast.children[0]
    if isinstance(ast, Token):
        if ast.type == "SIGNED_NUMBER": return int(ast.value)
        if ast.type == "BOOL": return ast.value == "true"
        if ast.type == "STRING": return ast.value[1:-1] # strip quotation marks
    elif isinstance(ast, Tree):
        if ast.data == "list":
            return [convert_value(val) for val in ast.children]

def convert_param(ast, file):
    doc = convert_docstring(ast.children[0])
    name = ast.children[1]
    opt = convert_opt(ast.children[2])
    type_ = convert_type(ast.children[3], file)

    location = Location(file, name.line, name.column, len(name.value))

    return SusField(location, doc, name.value, type_, opt)

def convert_timeout(val):
    num, mul = re.match("(\d+)(\w+)", val).groups()
    num = int(num)
    return num * {
        "ms":                      1,
        "s":                    1000,
        "m":               60 * 1000,
        "h":             3600 * 1000,
        "d":        24 * 3600 * 1000,
        "mo":  30 * 24 * 3600 * 1000,
        "y":  356 * 24 * 3600 * 1000,
    }[mul]

def convert_method(ast, file):
    static = None
    if ast.data == "static_method": static = True
    elif ast.data == "normal_method": static = False

    doc = convert_docstring(ast.children[0])
    name = ast.children[1]
    value = int(ast.children[2].value)

    params, returns, errors, confirmations, states, rate_limit = [], [], [], [], [], None
    for directive in ast.children[3:]:
        if not directive: continue
        if directive.data == "method_param":
            params.append(convert_param(directive, file))
        elif directive.data == "returns":
            for p in directive.children:
                returns.append(convert_param(p, file))
        elif directive.data in ["errors", "confirmations"]:
            lst = {"errors": errors, "confirmations": confirmations}[directive.data]
            for e in directive.children:
                if e.value in lst:
                    SourceError([Location(file, e.line, e.column, len(e.value))],
                        f"Duplicate member \"{e.value}\" for this directive").print_warn()
                lst.append(e.value)
        elif directive.data == "rate_limit":
            amount = int(directive.children[0].value)
            window = convert_timeout(directive.children[1].value)
            rate_limit = (amount, window)

    return SusMethod(Location(file, name.line, name.column, len(name.value)), doc, static, name.value,
        value, params, returns, errors, confirmations, rate_limit)

def convert_ast(ast, file):
    if ast.data in ["enum", "bitfield"]:
        constructor = SusEnum if ast.data == "enum" else SusBitfield
        doc = convert_docstring(ast.children[0])
        size = int(ast.children[1].value)
        name = ast.children[2]
        return constructor(Location(file, name.line, name.column, len(name)), doc,
            name.value, size, convert_ebf_members(ast.children[3:], file))

    elif ast.data == "entity":
        doc = convert_docstring(ast.children[0])
        name = ast.children[1]
        value = ast.children[2]

        directives = ast.children[3:]
        fields, methods = [], []
        for directive in directives:
            if directive.data == "entity_field":
                doc = convert_docstring(directive.children[0])
                identifier = directive.children[1]
                opt = convert_opt(directive.children[2])
                type_ = convert_type(directive.children[3], file)
                location = Location(file, identifier.line, identifier.column, len(identifier.value))

                fields.append(SusField(location, doc, identifier.value, type_, opt))

            if directive.data.endswith("method"):
                methods.append(convert_method(directive, file))

        return SusEntity(Location(file, name.line, name.column, len(name.value)), doc,
            name.value, int(value.value), fields, methods)

    elif ast.data == "global_method":
        return convert_method(ast, file)

    elif ast.data == "confirmation":
        doc = convert_docstring(ast.children[0])
        name = ast.children[1]
        value = int(ast.children[2].value)
        req = ast.children[3]
        resp = ast.children[4]

        req = [convert_param(par, file) for par in req.children if par]
        resp = [convert_param(par, file) for par in resp.children if par]

        return SusConfirmation(Location(file, name.line, name.column, len(name)), doc, name.value, value,
            req, resp)

    elif ast.data == "compound":
        doc = convert_docstring(ast.children[0])
        name = ast.children[1]

        directives = ast.children[2:]
        fields, methods = [], []
        for directive in directives:
            if not directive: continue
            doc = convert_docstring(directive.children[0])
            identifier = directive.children[1]
            opt = convert_opt(directive.children[2])
            type_ = convert_type(directive.children[3], file)
            location = Location(file, identifier.line, identifier.column, len(identifier.value))

            fields.append(SusField(location, doc, identifier.value, type_, opt))

        return SusCompound(Location(file, name.line, name.column, len(name.value)), doc,
            name.value, fields)



from colorama import Fore
from textwrap import dedent
from dataclasses import dataclass
from os.path import basename
from typing import *
from enum import Enum

from . import log

@dataclass
class Location():
    file: Any
    line: int
    col: int
    dur: int
    def __repr__(self):
        return f"{basename(self.file.path)}:{self.line}:{self.col}({self.dur})"
    def __hash__(self) -> int:
        return hash(self.file.source) + self.line + self.col + self.dur

class DiagLevel(Enum):
    ERROR = 1
    WARN = 2
    INFO = 3

@dataclass
class Diagnostic():
    locations: list[Location]
    level: DiagLevel
    message: str

class SusError(Exception):
    pass

SINGLE_LINE_ERRORS = False
class SourceError(SusError):
    def __init__(self, diag: Diagnostic):
        self.diag = diag
        self.accent = {
            DiagLevel.ERROR: Fore.RED,
            DiagLevel.WARN: Fore.YELLOW,
            DiagLevel.INFO: Fore.BLUE,
        }[diag.level]

    def print(self):
        function = {
            DiagLevel.ERROR: log.error,
            DiagLevel.WARN: log.warn,
            DiagLevel.INFO: log.info,
        }[self.diag.level]
        function(str(self))

    def __str__(self):
        if SINGLE_LINE_ERRORS:
            location = self.diag.locations[0]
            location = f"{location.file.path}:{location.line}:{location.col}"
            return f"{location}: {self.text}"

        error = Fore.WHITE + ("" if len(self.diag.locations) in [0, 1] else "multiple locations: ")

        for loc in self.diag.locations:
            padding = " " * (len(str(loc.line)) + 4 + loc.col - 1)
            squiggly = "~" * loc.dur if loc.dur > 0 else "^"
            src_line = loc.file.source.split('\n')[loc.line - 1]

            # get the inclusion path
            inclusion = Fore.LIGHTBLACK_EX
            cur = loc.file
            while cur.parent != None:
                cur = cur.parent
                inclusion += f" included from {cur.path}"

            error += dedent(f"""\
            {Fore.LIGHTBLACK_EX}in {Fore.WHITE}{loc.file.path}:{loc.line}:{loc.col}{inclusion}:
            {Fore.LIGHTBLACK_EX}{loc.line}  |{Fore.WHITE} {log.highlight_syntax(src_line)}
            {padding}{self.accent}{squiggly}
            """)

        error += f"{self.accent}{self.diag.message}{Fore.RESET}\n"
        return error.strip("\n")

class OutputError(SusError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class SearchError(SusError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

with open("requirements.txt", "r") as f:
    REQUIREMENTS = f.read().splitlines()

setup(
    name="susc",
    version="1.3.0",
    description="AMOGUS SUS description language compiler",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/amogus-api/susc",
    author="portasynthinca3",
    author_email="hi@portasynthinca3.me",
    license="BSD 3-Clause",
    packages=["susc"],
    keywords=["api", "protocol", "language"],
    install_requires=REQUIREMENTS,
    include_package_data=True,
    entry_points = {
        'console_scripts': ['susc=susc.__main__:main'],
    }
)



from colorama import Fore, Back
from sys import stderr
import re, lark

VERBOSE = False
ALL_STDERR = False
SUS_COLORS = {
    "(\".*\")|('.*')": Fore.LIGHTGREEN_EX,
    "/.+/[ims]{0,3}": Fore.LIGHTCYAN_EX,
    "#.*$": Fore.LIGHTBLACK_EX,
    "\\b([0-9]+|false|true|([0-9]+(y|mo|d|h|m|s|ms)))\\b": Fore.LIGHTYELLOW_EX,
    "\\b(enum|bitfield|confirmation|entity|opt|cache|hard|soft|method|staticmethod|globalmethod|compound|request|response|returns|errors|confirmations|states|ratelimit|every)\\b": Fore.MAGENTA,
    "[.,;:]": Fore.LIGHTBLUE_EX,
    "\\b[a-z][a-z_]*\\b": Fore.WHITE,
    "\\b[A-Z][A-Za-z]*\\b": Fore.YELLOW,
}

def log(back, fore, prefix, text, file):
    if ALL_STDERR:
        file = stderr
    print(f"{back}{fore} {prefix} {Back.RESET}{Fore.WHITE} {text}", file=file)

def error(text):
    log(Back.RED, Fore.BLACK, "ERR!", text, stderr)
def warn(text):
    log(Back.YELLOW, Fore.BLACK, "WARN", text, stderr)
def info(text):
    log(Back.BLUE, Fore.BLACK, "INFO", text, None)
def done(text):
    log(Back.GREEN, Fore.BLACK, "DONE", text, None)
def verbose(text, label=None):
    label = "DEBUG" + (":" + label if label else "")
    if VERBOSE:
        log(Back.LIGHTBLACK_EX, Fore.WHITE, label, text, None)

def highlight_syntax(line, colors=SUS_COLORS):
    if colors is SUS_COLORS:
        if line.startswith("include"):
            return Fore.MAGENTA + "include" + Fore.LIGHTGREEN_EX + line[7:]
        if line.startswith("set"):
            return Fore.MAGENTA + "set" + Fore.LIGHTGREEN_EX + line[3:]

    color_tokens = []
    for regex, color in colors.items():
        for entry in re.finditer(regex, line):
            start, end = entry.span()
            span = range(start, end)
            # check if something is already in these bounds
            flag = True
            for ex_span, _ in color_tokens:
                if len(set(ex_span).intersection(span)) != 0:
                    flag = False
                    break
            if not flag:
                continue

            color_tokens.append((span, color))
    color_tokens = sorted(color_tokens, key=lambda t: t[0][0])

    # highlight ranges while maintaining an offset
    offs = 0
    for span, color in color_tokens:
        start, end = span[0] + offs, span[-1] + 1 + offs
        replacement = color + line[start:end] + Fore.WHITE
        line = line[:start] + replacement + line[end:]
        offs += len(color) + len(Fore.WHITE)

    return line

def highlight_ast(ast):
    if not VERBOSE:
        return str(ast)
    if isinstance(ast, lark.Tree):
        return f"{Fore.YELLOW}{ast.data}{Fore.WHITE}[{', '.join([highlight_ast(c) for c in ast.children])}{Fore.WHITE}]"
    elif isinstance(ast, lark.Token):
        return f"{Fore.CYAN}{ast.type}{Fore.WHITE}{Fore.GREEN}'{ast.value}'{Fore.WHITE}"
    elif ast is None:
        return Fore.LIGHTBLACK_EX + "None" + Fore.WHITE

def highlight_thing(thing):
    thing = str(thing)
    if not VERBOSE:
        return thing
    thing = highlight_syntax(thing, {
        "'[^,']+'": Fore.GREEN,
        "location=[^:]+:[0-9]+:[0-9]+\([0-9]+\)": Fore.LIGHTBLACK_EX,
        "[a-z][a-z_]+=": Fore.CYAN,
        "\\bSus[A-Za-z]+\\b": Fore.YELLOW,
        "\\b(None|True|False|[0-9]+)\\b": Fore.LIGHTYELLOW_EX,
        "[.,;:]": Fore.LIGHTBLUE_EX,
    })
    
    return thing



from .things import *
from . import log
from .exceptions import *

MAGIC_IDENTIFIERS = ["Entity"]

def validate_fields(identifiers: List[str], field_sets: List[List[SusField]]) -> None:
    log.verbose("Validating fields", "linker")
    diag = []

    for fields in field_sets:
        for f1 in fields:
            # optional values mod 256
            if f1.optional != None and f1.optional >= 256:
                diag.append(Diagnostic([f1.location], DiagLevel.WARN,
                    f"Optional value '{f1.optional}' will be taken mod 256 ('{f1.optional % 256}')"))
                f1.optional %= 256

            # the fields within one set shouldn't have matching names
            equal = [f2 for f2 in fields if f2.name == f1.name]
            if len(equal) > 1:
                diag.append(Diagnostic([f.location for f in equal], DiagLevel.ERROR,
                    f"Multiple fields with matching names '{f1.name}'"))
            
            # or values
            equal = [f2 for f2 in fields if f2.optional == f1.optional and f1.optional != None]
            if len(equal) > 1:
                diag.append(Diagnostic([f.location for f in equal], DiagLevel.ERROR,
                    f"Multiple fields with matching opt() values '{f1.optional}'"))

            # validate the type
            type_err = f1.type_.find_errors(identifiers)
            if type_err != None:
                diag.append(Diagnostic([f1.type_.location], DiagLevel.ERROR, type_err))

    return diag

def combine(things: List[SusThing]) -> Tuple[list[SusThing], list[Diagnostic]]:
    log.verbose(f"Combining {len(things)} total definitions", "linker")
    diag = []
    ignore, out = [], []

    for thing1 in things:
        if thing1.name in ignore:
            continue

        # find things with same name
        with_matching_name = [thing2 for thing2 in things if thing2.name == thing1.name]

        if len(with_matching_name) > 1:
            # diagnose redefinitions
            fine = True
            for thing2 in with_matching_name:
                if not isinstance(thing1, (SusEnum, SusBitfield)):
                    diag.append(Diagnostic([t.location for t in with_matching_name], DiagLevel.ERROR,
                        f"Redefinition of '{thing1.name}' (only enums and bitfields can be combined)"))
                    fine = False

            if fine:
                # match sizes
                if not all([t.size == thing1.size for t in with_matching_name]):
                    diag.append(Diagnostic([t.location for t in with_matching_name],
                        DiagLevel.ERROR, "Can't combine things of different sizes"))

                # record members of enums and mitfields
                opt_members = []
                for t in with_matching_name:
                    opt_members += t.members

                # finally, combine all members
                constructor = SusEnum if isinstance(thing1, SusEnum) else SusBitfield
                doc = (thing1.docstring or "") + "\n" + (thing2.docstring or "")
                if doc == "\n": doc = None
                new_thing = constructor(thing1.location, doc, thing1.name, thing1.size, opt_members)
                out.append(new_thing)
                log.verbose(f"{Fore.LIGHTBLACK_EX}Combined {constructor.__name__[3:].lower()} {Fore.WHITE}{thing1.name}{Fore.LIGHTBLACK_EX} members across {len(with_matching_name)} definitions: {Fore.WHITE}{new_thing}", "linker")
        else:
            out.append(thing1)
        ignore.append(thing1.name)

        # check numeric values
        if isinstance(thing1, (SusEnum, SusBitfield)):
            maximum = (thing1.size * 8 if isinstance(thing1, SusBitfield) else 256 ** thing1.size) - 1
            for member in thing1.members:
                if member.value > maximum:
                    diag.append(Diagnostic([member.location], DiagLevel.WARN,
                        f"Member value '{member.value}' overflow (max '{maximum}')"))

        if isinstance(thing1, (SusEntity, SusMethod)) and thing1.value > 127:
            diag.append(Diagnostic([thing1.location], DiagLevel.WARN, f"Value '{thing1.value}' overflow (max '127')"))

        if isinstance(thing1, SusConfirmation) and thing1.value > 15:
            diag.append(Diagnostic([thing1.location], DiagLevel.WARN, f"Value '{thing1.value}' overflow (max '15')"))

    log.verbose(f"Combined {len(things)} definitions into {len(out)} things", "linker")
    return out, diag

def validate_method_meta(things: List[SusThing], method_sets: List[List[SusMethod]]) -> None:
    log.verbose("Validating method metdata", "linker")
    diag = []

    errors = [t for t in things if isinstance(t, SusEnum) and t.name == "ErrorCode"]
    confirmations = [t.name for t in things if isinstance(t, SusConfirmation)]

    for m_set in method_sets:
        for method in m_set:
            for conf in method.confirmations:
                if conf not in confirmations:
                    diag.append(Diagnostic([method.location], DiagLevel.ERROR, f"Undefined confirmation '{conf}'"))

            if len(method.errors) > 0 and len(errors) == 0:
                diag.append(Diagnostic([method.location], DiagLevel.WARN,
                    "No 'ErrorCode' enum defined. Include 'impostor.sus' or use a custom definition"))
                continue
            if len(method.errors) == 0:
                continue
            error_names = [m.name for m in errors[0].members]

            for err in method.errors:
                if err not in error_names:
                    diag.append(Diagnostic([method.location], DiagLevel.ERROR, f"Undefined error code '{err}'"))
    
    return diag

def validate_values(entities: List[SusEntity], method_sets: List[List[SusMethod]], confirmations: List[SusConfirmation]) -> None:
    log.verbose("Validating numeric values", "linker")
    diag = []

    for thing in entities:
        matching = [t for t in entities if t.value == thing.value]
        if len(matching) != 1:
            diag.append(Diagnostic([t.location for t in matching], DiagLevel.ERROR,
                f"Multiple entities with matching values '{thing.value}'"))

        # check ID field
        id_field = [f for f in thing.fields if f.name == "id"]
        if not id_field:
            diag.append(Diagnostic([thing.location], DiagLevel.WARN, f"No 'id' field"))
        else:
            id_field = id_field[0]
            t = id_field.type_
            if t.name != "Int" or len(t.args) != 1 or t.args[0] != 8:
                diag.append(Diagnostic([id_field.location], DiagLevel.WARN, f"The 'id' field is not an Int(8)"))

    for m_set in method_sets:
        for method in m_set:
            matching = [t for t in m_set if t.value == method.value]
            if len(matching) != 1:
                diag.append(Diagnostic([t.location for t in matching], DiagLevel.ERROR,
                    f"Multiple methods with matching values '{method.value}'"))

    for thing in confirmations:
        matching = [t for t in confirmations if t.value == thing.value]
        if len(matching) != 1:
            diag.append(Diagnostic([t.location for t in matching], DiagLevel.ERROR,
                f"Multiple confirmations with matching values '{thing.value}'"))
    
    return diag

def strip_docstrings(things: List[SusThing]) -> List[SusThing]:
    for thing in things:
        doc = thing.docstring
        thing.docstring = doc.strip() if doc else None
    return things

def run(things: List[SusThing]) -> List[SusThing]:
    log.verbose("Running linker", "linker")

    # get all identifiers that can be referenced
    identifiers = [t.name for t in things if not isinstance(t, SusMethod)] + MAGIC_IDENTIFIERS
    # get all possible fields: entity fields, method params and method return vals
    entities = [t for t in things if isinstance(t, SusEntity)]
    methods = [t for t in things if isinstance(t, SusMethod)]
    confirmations = [t for t in things if isinstance(t, SusConfirmation)]
    compounds = [t for t in things if isinstance(t, SusCompound)]
    for e in entities: methods += e.methods
    field_sets = [e.fields for e in entities]
    field_sets += [m.parameters for m in methods]
    field_sets += [m.returns for m in methods]
    field_sets += [m.req_parameters for m in confirmations]
    field_sets += [m.resp_parameters for m in confirmations]
    field_sets += [c.fields for c in compounds]
    # get all methods
    method_sets = [[t for t in things if isinstance(t, SusMethod)]]
    for e in entities:
        method_sets.append([m for m in e.methods if m.static])
        method_sets.append([m for m in e.methods if not m.static])

    # run substeps collecting diagnostics
    things, diag = combine(things)
    diag += validate_fields(identifiers, field_sets)
    diag += validate_method_meta(things, method_sets)
    diag += validate_values(entities, method_sets, confirmations)
    things = strip_docstrings(things)

    # sort by severity
    diag = sorted(diag, key=lambda d: d.level.value)

    # remove diagnostics in the same location
    deduplicated = []
    used_locs = set()
    for d in diag:
        for loc in d.locations:
            has_duplicates = True
            if loc not in used_locs:
                has_duplicates = False
                used_locs.add(loc)
        if not has_duplicates:
            deduplicated.append(d)

    return things, deduplicated